--- old/usr/src/uts/common/fs/zfs/arc.c	Tue Nov 11 06:35:52 2014
+++ new/usr/src/uts/common/fs/zfs/arc.c	Tue Nov 11 06:35:52 2014
@@ -2519,6 +2519,7 @@
 
 	arc_adapt(size, state);
 
+top:
 	/*
 	 * We have not yet reached cache maximum size,
 	 * just allocate a new buffer.
@@ -2525,15 +2526,24 @@
 	 */
 	if (!arc_evict_needed(type)) {
 		if (type == ARC_BUFC_METADATA) {
-			buf->b_data = zio_buf_alloc(size);
-			arc_space_consume(size, ARC_SPACE_DATA);
+			buf->b_data = zio_buf_alloc_canfail(size);
+			if (buf->b_data != NULL) {
+				arc_space_consume(size, ARC_SPACE_DATA);
+				goto out;
+			}
 		} else {
 			ASSERT(type == ARC_BUFC_DATA);
-			buf->b_data = zio_data_buf_alloc(size);
-			ARCSTAT_INCR(arcstat_data_size, size);
-			atomic_add_64(&arc_size, size);
+			buf->b_data = zio_data_buf_alloc_canfail(size);
+			if (buf->b_data != NULL) {
+				ARCSTAT_INCR(arcstat_data_size, size);
+				atomic_add_64(&arc_size, size);
+				goto out;
+			}
 		}
-		goto out;
+		/*
+		 * Memory allocation failed probably due to excessive
+		 * fragmentation, we need to evict regardless.
+		 */
 	}
 
 	/*
@@ -2556,16 +2566,8 @@
 		    mfu_space > arc_mfu->arcs_size) ? arc_mru : arc_mfu;
 	}
 	if ((buf->b_data = arc_evict(state, NULL, size, TRUE, type)) == NULL) {
-		if (type == ARC_BUFC_METADATA) {
-			buf->b_data = zio_buf_alloc(size);
-			arc_space_consume(size, ARC_SPACE_DATA);
-		} else {
-			ASSERT(type == ARC_BUFC_DATA);
-			buf->b_data = zio_data_buf_alloc(size);
-			ARCSTAT_INCR(arcstat_data_size, size);
-			atomic_add_64(&arc_size, size);
-		}
 		ARCSTAT_BUMP(arcstat_recycle_miss);
+		goto top;
 	}
 	ASSERT(buf->b_data != NULL);
 out:
